(herald "IKEv2 EAP" (algebra diffie-hellman))

;; This protocol descriptions follows the Scyther model found on the
;; Scyther models page. It appears to have a number of inaccuracies,
;; but we follow it closely in order to make a fair comparison for
;; performance results.

;; Its treatment of EAP payloads as simple data,
;; especially the repetition of the value eap is not very
;; realistic. Those payloads are meant to represent arbitrary EAP
;; payloads, and as such, the two eap payloads should be distinct at
;; the very least. Indeed, the RFC states that AUTH values after the
;; EAP exchange must contain the EAP AUTH payloads which is not done
;; here. It also repeats what appear to be mistakes surrounding SA1
;; and SA2.

(defmacro (prf x y) (hash "prf" x y))
(defmacro (kdf x) (hash "kdf" x))

(defmacro (hdr) (cat SPIi SPIr))

(defmacro (Zi) (exp Gr i))
(defmacro (Zr) (exp Gi r))

(defmacro (SKi) (kdf (cat ni nr (Zi) SPIi SPIr)))
(defmacro (SKr) (kdf (cat ni nr (Zr) SPIi SPIr)))

(defmacro (AUTHii) (enc SPIi "0" SA1 (exp (gen) i) ni nr (prf (SKi) I) (privk I)))
(defmacro (AUTHir) (enc SPIi "0" SA1 Gi ni nr (prf (SKr) I) (privk I)))
(defmacro (AUTHri) (enc SPIi SPIr SA1 Gr nr ni (prf (SKi) R) (privk R)))
(defmacro (AUTHrr) (enc SPIi SPIr SA1 (exp (gen) r) nr ni (prf (SKr) R) (privk R)))

(defprotocol ikev2-eap diffie-hellman
  (defrole roleI
    (vars (I R name) (SPIi SPIr SA1 SA2 ni nr TSi TSr data)
          (i expn) (Gr base) (eap eapok data))
    (trace
     (send (cat I R SPIi "0" SA1 (exp (gen) i) ni))
     (recv (cat R I (hdr) SA1 Gr nr))
     (send (cat I R (hdr) (enc I R SA2 TSi TSr (SKi))))
     (recv (cat R I (hdr) (enc R (AUTHri) eap (SKi))))
     (send (cat I R (hdr) (enc eap (SKi))))
     (recv (cat R I (hdr) (enc eapok (SKi))))
     (send (cat I R (hdr) (enc (AUTHii) (SKi))))
     (recv (cat R I (hdr) (enc (AUTHri) SA2 TSi TSr (SKi)))))
    (uniq-orig ni SPIi)
    (uniq-gen i))
  (defrole roleR
    (vars (I R name) (SPIi SPIr SA1 SA2 ni nr TSi TSr data)
          (r expn) (Gi base) (eap eapok data))
    (trace
     (recv (cat I R SPIi "0" SA1 Gi ni))
     (send (cat R I (hdr) SA1 (exp (gen) r) nr))
     (recv (cat I R (hdr) (enc I R SA2 TSi TSr (SKr))))
     (send (cat R I (hdr) (enc R (AUTHrr) eap (SKr))))
     (recv (cat I R (hdr) (enc eap (SKr))))
     (send (cat R I (hdr) (enc eapok (SKr))))
     (recv (cat I R (hdr) (enc (AUTHir) (SKr))))
     (send (cat R I (hdr) (enc (AUTHrr) SA2 TSi TSr (SKr)))))
    (uniq-orig nr SPIr)
    (uniq-gen r)
    (uniq-orig eap eapok)
    ))


;; Secrecy Goals
(defgoal ikev2-eap
  (forall ((I R name) (ni nr SPIi SPIr data) (Gr base) (i expn) (z z-0 node))
          (implies
           (and (p "roleI" 7 z)
                (p "roleI" "I" z I)
                (p "roleI" "R" z R)
                (p "roleI" "ni" z ni)
                (p "roleI" "nr" z nr)
                (p "roleI" "Gr" z Gr)
                (p "roleI" "i" z i)
                (p "roleI" "SPIi" z SPIi)
                (p "roleI" "SPIr" z SPIr)
                (p "" 0 z-0)
                (p "" "x" z-0 (SKi))
                (non (privk I))
                (non (privk R)))
           (false))))

(defgoal ikev2-eap
  (forall ((I R name) (ni nr SPIi SPIr data) (Gi base) (r expn) (z z-0 node))
          (implies
           (and (p "roleR" 7 z)
                (p "roleR" "I" z I)
                (p "roleR" "R" z R)
                (p "roleR" "ni" z ni)
                (p "roleR" "nr" z nr)
                (p "roleR" "Gi" z Gi)
                (p "roleR" "r" z r)
                (p "roleR" "SPIi" z SPIi)
                (p "roleR" "SPIr" z SPIr)
                (p "" 0 z-0)
                (p "" "x" z-0 (SKr))
                (non (privk I))
                (non (privk R)))
           (false))))

;; Non-injective Agreement Goals
(defgoal ikev2-eap
  (forall ((I R name) (ni nr SPIi SPIr SA1 SA2 TSi TSr eap data) (Gr base) (i expn) (z node))
          (implies
           (and (p "roleI" 7 z)
                (p "roleI" "I" z I)
                (p "roleI" "R" z R)
                (p "roleI" "ni" z ni)
                (p "roleI" "nr" z nr)
                (p "roleI" "Gr" z Gr)
                (p "roleI" "i" z i)
                (p "roleI" "SPIi" z SPIi)
                (p "roleI" "SPIr" z SPIr)
                (p "roleI" "SA1" z SA1)
                (p "roleI" "SA2" z SA2)
                (p "roleI" "TSi" z TSi)
                (p "roleI" "TSr" z TSr)
                (p "roleI" "eap" z eap)
                (non (privk I))
                (non (privk R)))
           (exists ((z-1 node) (Gi base) (r expn))
                  (and (p "roleR" 7 z-1)
                       (p "roleR" "I" z-1 I)
                       (p "roleR" "R" z-1 R)
                       (p "roleR" "ni" z-1 ni)
                       (p "roleR" "nr" z-1 nr)
                       (p "roleR" "Gi" z-1 Gi)
                       (p "roleR" "r" z-1 r)
                       (p "roleR" "SPIi" z-1 SPIi)
                       (p "roleR" "SPIr" z-1 SPIr)
                       (p "roleR" "SA1" z-1 SA1)
                       (p "roleR" "SA2" z-1 SA2)
                       (p "roleR" "TSi" z-1 TSi)
                       (p "roleR" "TSr" z-1 TSr)
                       (p "roleR" "eap" z-1 eap)
                       (= (exp Gr i) (exp Gi r)))))))

(defgoal ikev2-eap
  (forall ((I R name) (ni nr SPIi SPIr SA1 SA2 TSi TSr eap data) (Gi base) (r expn) (z node))
          (implies
           (and (p "roleR" 7 z)
                (p "roleR" "I" z I)
                (p "roleR" "R" z R)
                (p "roleR" "ni" z ni)
                (p "roleR" "nr" z nr)
                (p "roleR" "Gi" z Gi)
                (p "roleR" "r" z r)
                (p "roleR" "SPIi" z SPIi)
                (p "roleR" "SPIr" z SPIr)
                (p "roleR" "SA1" z SA1)
                (p "roleR" "SA2" z SA2)
                (p "roleR" "TSi" z TSi)
                (p "roleR" "TSr" z TSr)
                (p "roleR" "eap" z eap)
                (non (privk I))
                (non (privk R)))
           (exists ((z-1 node) (Gr base) (i expn))
                  (and (p "roleI" 6 z-1)
                       (p "roleI" "I" z-1 I)
                       (p "roleI" "R" z-1 R)
                       (p "roleI" "ni" z-1 ni)
                       (p "roleI" "nr" z-1 nr)
                       (p "roleI" "Gr" z-1 Gr)
                       (p "roleI" "i" z-1 i)
                       (p "roleI" "SPIi" z-1 SPIi)
                       (p "roleI" "SPIr" z-1 SPIr)
                       (p "roleI" "SA1" z-1 SA1)
                       (p "roleI" "SA2" z-1 SA2)
                       (p "roleI" "TSi" z-1 TSi)
                       (p "roleI" "TSr" z-1 TSr)
                       (p "roleI" "eap" z-1 eap)
                       (= (exp Gr i) (exp Gi r)))))))



           
    
